import math
import warnings
from collections import Counter

import SimpleITK as sitk
import itk
import networkx as nx
import numpy as np
import vtk
from vtk.util.numpy_support import numpy_to_vtk

from .helpers import slice_normalize
from .chunky import Sparse
from .vtk_utils import add_np_to_vti
from .geometry.extract import triangles

dilate_filter = sitk.BinaryDilateImageFilter()
erode_filter = sitk.BinaryErodeImageFilter()
closing_filter = sitk.BinaryMorphologicalClosingImageFilter()
opening_filter = sitk.BinaryMorphologicalOpeningImageFilter()
connected_compontent_filter = sitk.ConnectedComponentImageFilter()


def min_dtype(t):
    try:
        t = t.type
    except:
        pass

    if issubclass(t, np.inexact):
        return np.finfo(t).min
    else:
        return np.iinfo(t).min


def max_dtype(t):
    try:
        t = t.type
    except:
        pass

    if issubclass(t, np.inexact):
        return np.finfo(t).max
    else:
        return np.iinfo(t).max


def _unique_pairs(a):
    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    return np.unique(b).view(a.dtype).reshape(-1, a.shape[1])


def unique(sparse, return_counts=False, multiprocesses=1):
    """ see: https://docs.scipy.org/doc/numpy/reference/generated/numpy.unique.html """
    def process(data, prev):
        if return_counts:
            values, counts = np.unique(data, return_counts=True)
            x = dict(zip(values, counts))

            return data, dict(Counter(x) + Counter(prev))
        else:
            return data, prev | set(np.unique(data))

    result = sparse.run(process, prev={} if return_counts else set(), multiprocesses=multiprocesses)

    result_merged = {} if return_counts else set()
    for process in result:
        if return_counts:
            result_merged = dict(Counter(result_merged) + Counter(process))
        else:
            result_merged |= process
    result = result_merged
    return result


def downscale(sparse, stride=(2, 2, 2)):
    ret = Sparse.empty_like(sparse,
                            chunks=(sparse.chunks[0] // stride[0],
                                    sparse.chunks[1] // stride[1],
                                    sparse.chunks[2] // stride[2]),
                            shape=(sparse.shape[0] // stride[0],
                                   sparse.shape[1] // stride[1],
                                   sparse.shape[2] // stride[2]),
                            spacing=(sparse.spacing[0] * stride[0],
                                     sparse.spacing[1] * stride[1],
                                     sparse.spacing[2] * stride[2]))

    for k in sparse._grid.keys():
        value = sparse._grid[k][0] if sparse._grid[k].shape == (1,) else sparse._grid[k][::stride[0], ::stride[1],
                                                                         ::stride[2]]
        ret.set_chunk(k, value)

    return ret


def sum(sparse, multiprocesses=1):
    return np.sum(sparse.run(lambda data, prev: (data, prev + np.sum(data)), prev=0, multiprocesses=multiprocesses))


def downsample(sparse, stride=(3, 3, 3)):
    sparse_copy = Sparse.empty_like(sparse, chunks=np.array(stride) * 16)
    sparse_copy.copy_from(sparse)

    sparse_downscaled = Sparse(np.array(sparse.shape) // np.array(stride), dtype=np.float32)
    sparse_downscaled.origin = sparse.origin
    sparse_downscaled.spacing = np.array(sparse.spacing) * np.array(stride)

    def reduce(chunk, prev):
        accu = np.zeros((16,) * 3, dtype=np.float32)

        for i in range(stride[0]):
            for j in range(stride[1]):
                for k in range(stride[2]):
                    accu += chunk[i::stride[0], j::stride[1], k::stride[2]]

        coord = sparse_copy._chunk_to_global_coord(chunk.idx, (0, 0, 0)) // stride

        prev[coord[0]:coord[0] + 16,
        coord[1]:coord[1] + 16,
        coord[2]:coord[2] + 16] = accu / (stride[0] * stride[1] * stride[2])

        return chunk, prev

    return sparse_copy.run(reduce, (0, 0, 0), sparse_downscaled)


def sum_slice(sparse, z, y, x):
    def process(idx, gs, z, y, x, prev):
        chunk_size = abs(abs(x.stop - x.start) // x.step) * \
                     abs(abs(y.stop - y.start) // y.step) * \
                     abs(abs(z.stop - z.start) // z.step)

        if idx in sparse._grid.keys():
            if sparse._grid[idx].shape == (1,):
                prev += sparse._grid[idx][0] * chunk_size
            else:
                prev += np.sum(sparse._grid[idx][z, y, x])
        else:
            prev += sparse.fill_value * chunk_size

        return prev

    z, y, x = slice_normalize((z, y, x), sparse.shape)

    return sparse.raw_run(process,
                          (z.start, y.start, x.start),
                          (z.stop, y.stop, x.stop),
                          step=(z.step, y.step, x.step),
                          prev=0)


def where(sparse, func):
    ret_i, ret_j, ret_k = [], [], []

    for key in sorted(sparse._grid.keys()):
        i, j, k = np.where(func(sparse.get_chunk(key)))

        offset = sparse._chunk_to_global_coord(key, (0, 0, 0))
        ret_i.append(i + offset[0])
        ret_j.append(j + offset[1])
        ret_k.append(k + offset[2])

    return np.hstack(ret_i), np.hstack(ret_j), np.hstack(ret_k)


def max(sparse, multiprocesses=1):
    f = np.max
    return f(sparse.run(
                        lambda data, prev: (data, f([prev, f(data)])),
                        prev=min_dtype(sparse.dtype),
                        multiprocesses=multiprocesses,
                        )
            )


def nanmax(sparse, multiprocesses=1):
    f = np.nanmax
    return f(sparse.run(
                        lambda data, prev: (data, f([prev, f(data)])),
                        prev=min_dtype(sparse.dtype),
                        multiprocesses=multiprocesses,
                        )
            )


def max_slice(sparse, z, y, x):
    z, y, x = slice_normalize((z, y, x), sparse.shape)

    return sparse.raw_run(lambda idx, gs, z, y, x, prev: np.max([prev, np.max(sparse.get_chunk(idx)[x, y, x])]),
                          (z.start, y.start, x.start),
                          (z.stop, y.stop, x.stop),
                          step=(z.step, y.step, x.step),
                          prev=-math.inf)


def min(sparse, multiprocesses=1):
    f = np.min
    return f(sparse.run(
                        lambda data, prev: (data, f([prev, f(data)])),
                        prev=max_dtype(sparse.dtype),
                        multiprocesses=multiprocesses,
                        )
            )


def nanmin(sparse, multiprocesses=1):
    f = np.nanmin
    return f(sparse.run(
                        lambda data, prev: (data, f([prev, f(data)])),
                        prev=max_dtype(sparse.dtype),
                        multiprocesses=multiprocesses,
                        )
            )


def min_slice(sparse, z, y, x):
    z, y, x = slice_normalize((z, y, x), sparse.shape)

    return sparse.raw_run(lambda idx, gs, z, y, x, prev: np.min([prev, np.min(sparse.get_chunk(idx)[x, y, x])]),
                          (z.start, y.start, x.start),
                          (z.stop, y.stop, x.stop),
                          step=(z.step, y.step, x.step),
                          prev=math.inf)


def mul(sparse_a, sparse_b):
    return sparse_a.run_multivariate(lambda a, b: a * b[0], [sparse_b])


def add(sparse_a, sparse_b):
    return sparse_a.run_multivariate(lambda a, b: a + b[0], [sparse_b])


def subtract(sparse_a, sparse_b):
    return sparse_a.run_multivariate(lambda a, b: a - b[0], [sparse_b])


def add_scalar(sparse_a, val, multiprocesses=1):
    sparse_a.run(lambda data, prev: (data + val, prev), multiprocesses=multiprocesses)


def dilate(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """ see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm """
    dilate_filter.SetKernelRadius(kernel_radius)
    dilate_filter.SetForegroundValue(foreground_value)
    sparse.run(lambda data, prev:
               (sitk.GetArrayFromImage(
                   dilate_filter.Execute(
                       sitk.GetImageFromArray(data)
                   )
               ), prev), envelope=kernel_radius, multiprocesses=multiprocesses)


def erode(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """ see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm """
    erode_filter.SetKernelRadius(kernel_radius)
    erode_filter.SetForegroundValue(foreground_value)
    sparse.run(lambda data, prev:
               (sitk.GetArrayFromImage(
                   erode_filter.Execute(
                       sitk.GetImageFromArray(data)
                   )
               ), prev), envelope=kernel_radius, skip_neighbours=True, multiprocesses=multiprocesses)


def closing(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """ see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm """
    closing_filter.SetKernelRadius(kernel_radius)
    closing_filter.SetForegroundValue(foreground_value)
    sparse.run(lambda data, prev:
               (sitk.GetArrayFromImage(
                   closing_filter.Execute(
                       sitk.GetImageFromArray(data)
                   )
               ), prev), envelope=kernel_radius, multiprocesses=multiprocesses)


def opening(sparse, kernel_radius, foreground_value=1, multiprocesses=1):
    """ see: http://homepages.inf.ed.ac.uk/rbf/HIPR2/morops.htm """
    opening_filter.SetKernelRadius(kernel_radius)
    opening_filter.SetForegroundValue(foreground_value)
    sparse.run(lambda data, prev: (sitk.GetArrayFromImage(
        opening_filter.Execute(
            sitk.GetImageFromArray(data)
        )
    ), prev), envelope=kernel_radius, skip_neighbours=True, multiprocesses=multiprocesses)


def thinning(sparse, envelope, multiprocesses=1):
    """
    1 pixel-thin wire skeletonization
    """
    sparse.run(lambda data, prev: (itk.GetArrayFromImage(
        itk.BinaryThinningImageFilter3D.New(
            itk.GetImageFromArray(data)
        )
    ), prev), envelope=envelope, skip_neighbours=True, multiprocesses=multiprocesses)


def thinning_diameter(sparse, envelope, multiprocesses=2):
    """
    Computes twice the shortest distance to the outter shell along the medial axis of the object
    (i.e. diameter of the local maximal fitting sphere)
    """
    if multiprocesses == 1:
        warnings.warn(
            'Disabled multiprocessing creates a pool of threads which does not get deallocated. '
            'Subsequent calls with multiprocessing **enabled** will be locked indefinitely.'
        )
    sparse.run(lambda data, prev: (itk.GetArrayFromImage(
        itk.MedialThicknessImageFilter3D.New(
            itk.GetImageFromArray(data)
        )
    ), prev), envelope=envelope, skip_neighbours=True, multiprocesses=multiprocesses)


def to_indices_value(sparse):
    coords = []

    for k in sparse._grid.keys():
        for val in np.unique(sparse._grid[k]):
            if val != sparse.fill_value:
                z, y, x = np.where(sparse._grid[k] == val)
                offset = sparse._chunk_to_global_coord(k, (0, 0, 0))
                x += offset[2]
                y += offset[1]
                z += offset[0]
                coords.append(np.dstack((z, y, x, np.full(x.shape, val))).reshape(-1, 4))

    return np.vstack(coords)


def label(sparse, multiprocesses=1, fully_connected=False):
    """
    Label the objects in a binary image.
    ConnectedComponentImageFilter labels the objects in a binary image (non-zero pixels are considered to be objects, zero-valued pixels are considered to be background). Each distinct object is assigned a unique label. The final object labels start with 1 and are consecutive.
    :param sparse: Sparse instance
    :param multiprocesses: number of processes
    :param fully_connected: whether the connected components are defined strictly by face connectivity or by face+edge+vertex connectivity. Set `True` for objects that are 1 pixel wide.
    """
    if multiprocesses != 1:
        raise NotImplementedError()

    if sparse.dtype != np.uint32:
        warnings.warn('label in most cases require uint32 data to work properly')
        
    connected_compontent_filter.SetFullyConnected(fully_connected)

    G = nx.Graph()

    def process(a, b):
        mask = np.dstack([a, b])
        relations = _unique_pairs(mask[np.sum(mask != 0, axis=2) == 2])

        for r in relations:
            G.add_edge(r[0], r[1])

    def initialize(data, prev):
        result = sitk.GetArrayFromImage(
            connected_compontent_filter.Execute(
                sitk.GetImageFromArray(data)
            )
        )

        obj_count = connected_compontent_filter.GetObjectCount()
        result[result > 0] += prev[1]
        prev[0].append(obj_count)

        return result, (prev[0], prev[1] + obj_count)

    component_count_per_chunk, component_sum = sparse.run(initialize, prev=([], 0), multiprocesses=multiprocesses)[0]

    for k in sparse._grid.keys():
        # Z axis
        nk = (k[0] - 1, k[1], k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[0, :, :], sparse.get_chunk(nk)[-1, :, :])

        nk = (k[0] + 1, k[1], k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[-1, :, :], sparse.get_chunk(nk)[0, :, :])

        # Y axis
        nk = (k[0], k[1] - 1, k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, 0, :], sparse.get_chunk(nk)[:, -1, :])

        nk = (k[0], k[1] + 1, k[2])
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, -1, :], sparse.get_chunk(nk)[:, 0, :])

        # X axis
        nk = (k[0], k[1], k[2] - 1)
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, :, 0], sparse.get_chunk(nk)[:, :, -1])

        nk = (k[0], k[1], k[2] + 1)
        if nk in sparse._grid.keys():
            process(sparse.get_chunk(k)[:, :, -1], sparse.get_chunk(nk)[:, :, 0])

    cutted_G = list(G.subgraph(c) for c in nx.connected_components(G))
    val_map = {v: k + 1 for k, g in enumerate(cutted_G) for v in g}

    for idx, k in enumerate(set(range(1, component_sum + 1)) - val_map.keys()):
        val_map[k] = idx + len(cutted_G) + 1

    add_scalar(sparse, component_sum)
    component_index = 1

    for k, component_count in zip(sorted(sparse._grid.keys()), component_count_per_chunk):
        chunk = sparse.get_chunk(k)

        for _ in range(component_count):
            if component_index in val_map:
                chunk[chunk == (component_index + component_sum)] = val_map[component_index]

            component_index += 1

        chunk[chunk >= component_sum] -= component_sum

        sparse.set_chunk(k, chunk)


def contour(sparse, values, envelope=(1,1,1), quantize_points_factor=0.0, multiprocesses=1):
    """
    Generate isosurface(s) from a Sparse volume
    """
    def chunk_contour(data, prev, *args):
        envelope, values = args

        vti = vtk.vtkImageData()
        vti.SetOrigin(data.origin)
        vti.SetDimensions(np.array(data.shape) - np.array(envelope))
        vti.SetSpacing(data.spacing)
        series_name = "sparse"
        add_np_to_vti(vti, data[:-envelope[0], :-envelope[1], :-envelope[2]], series_name)
        vti.GetPointData().SetActiveScalars(series_name)
        contour = vtk.vtkImageMarchingCubes()
        contour.SetInputData(vti)
        contour.SetNumberOfContours(values.shape[0])
        for i, v in enumerate(values):
            contour.SetValue(i, v)
        contour.ComputeScalarsOn()
        contour.Update()

        writer = vtk.vtkPolyDataWriter()
        writer.SetInputData(contour.GetOutput())
        writer.WriteToOutputStringOn()
        writer.Update()
        prev.append(writer.GetOutputString())

        return data, prev

    out = sparse.run(chunk_contour,
                     envelope,
                     np.array([values]).flatten(),
                     envelope=envelope,
                     multiprocesses=multiprocesses,
                     prev=[],
              )

    appendFilter = vtk.vtkAppendPolyData()
    for contour in out:
        for c in contour:
            reader = vtk.vtkPolyDataReader()
            reader.ReadFromInputStringOn()
            reader.SetInputString(c)
            reader.Update()
            appendFilter.AddInputData(reader.GetOutput())

    quantize = vtk.vtkQuantizePolyDataPoints()
    quantize.SetQFactor(quantize_points_factor)

    quantize.SetInputConnection(appendFilter.GetOutputPort())
    quantize.Update()

    return triangles(quantize.GetOutput())



import os.path
import logging

import vtk
from vtk.util import numpy_support
import numpy as np

logging.debug(("VTK version:", vtk.VTK_VERSION))


# region VTK I/O

# Great explanation of VTK file and data formats are in
# "VTK Textbook" and "VTK User Guide" chapter 12: "Readers" and 19: "VTK File Formats".


def read_vtk(path, return_reader=False):
    path = os.path.abspath(path)
    logging.debug('Reading: "{}".'.format(path))

    _, file_extension = os.path.splitext(path)

    if file_extension.endswith(".vti"):
        reader = vtk.vtkXMLImageDataReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".stl"):
        reader = vtk.vtkSTLReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".vtp"):
        reader = vtk.vtkXMLPolyDataReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".vtu"):
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".mhd"):
        reader = vtk.vtkMetaImageReader()
        reader.SetFileName(path)

    elif file_extension.endswith(".vtk"):
        reader = vtk.vtkUnstructuredGridReader()
        reader.SetFileName(path)
        reader.ReadAllVectorsOn()
        reader.ReadAllScalarsOn()
    else:
        raise Exception("Unsupported file extension.")

    reader.Update()

    if not return_reader:
        return reader.GetOutput()
    return reader.GetOutput(), reader


def save_vti(vti, path):
    _save(vti, path, vtk.vtkXMLImageDataWriter)


def save_vtp(vtp, path):
    _save(vtp, path, vtk.vtkXMLPolyDataWriter)


def save_vtu(vtu, path):
    _save(vtu, path, vtk.vtkXMLUnstructuredGridWriter)


def save_stl(vtp, path):
    _save(vtp, path, vtk.vtkSTLWriter)


def _save(input_data, path, writer_type):
    path = os.path.abspath(path)
    logging.debug('Saving: "{}".'.format(path))
    writer = writer_type()
    writer.SetFileName(path)
    writer.SetInputData(input_data)
    writer.Write()

# endregion

# region Conversions


def add_np_to_vti(vti, arr_np, arr_name, arr_type=None):
    arr = numpy_support.numpy_to_vtk(num_array=arr_np.ravel(), deep=True, array_type=arr_type)
    arr.SetName(arr_name)
    vti.GetPointData().AddArray(arr)


def vti_to_np(vti, array, components=1):
    x, y, z = vti.GetDimensions()
    arr_np = numpy_support.vtk_to_numpy(vti.GetPointData().GetArray(array))
    if components == 1:
        return arr_np.reshape(z, y, x)
    else:
        return arr_np.reshape(z, y, x, components)



def vtp_to_np(vtp, arrays):
    if vtp.GetPolys().GetMaxCellSize() > 3:
        cut_triangles = vtk.vtkTriangleFilter()
        cut_triangles.SetInputData(vtp)
        cut_triangles.Update()
        vtp = cut_triangles.GetOutput()

    result = [
        ('vertices', numpy_support.vtk_to_numpy(vtp.GetPoints().GetData()).astype(np.float32)),
        ('triangles', numpy_support.vtk_to_numpy(vtp.GetPolys().GetData()).reshape(-1, 4)[:, 1:4]),
    ]
    result.extend([(arr, numpy_support.vtk_to_numpy(vtp.GetPointData().GetArray(arr))) for arr in arrays])
    return dict(result)


def vti_to_string(vti):
    writer = vtk.vtkXMLImageDataWriter()
    writer.SetInputData(vti)
    writer.WriteToOutputStringOn()  # must be before Update call
    writer.Update()
    return writer.GetOutputString()


def vtp_to_string(vtp):
    polydata_writer = vtk.vtkPolyDataWriter()
    polydata_writer.WriteToOutputStringOn()
    polydata_writer.SetInputData(vtp)
    polydata_writer.Update()
    return polydata_writer.GetOutputString()


def string_to_vtp(s):
    reader = vtk.vtkPolyDataReader()
    reader.ReadFromInputStringOn()
    reader.SetInputString(s)
    reader.Update()
    return reader.GetOutput()


def sailfish_vti_to_npy(vti_file, verbose=False, rho_name='rho', v_name='v'):
    """
    Read vti files from sailfish and return
    gets only rho and v
    """
    vti, reader = read_vtk(vti_file, return_reader=True)
    # info = reader.GetInformation()

    field_names = [reader.GetPointArrayName(i) for i in range(reader.GetNumberOfPointArrays())]
    logging.getLogger(vti_file).debug(("fields:", field_names))
    assert (rho_name in field_names and v_name in field_names)

    data_rho = vti_to_np(vti, rho_name)
    data_v = vti_to_np(vti, v_name, components=3)
    data_v = np.rollaxis(data_v, 3, 0)
    # alt.  data = data.transpose(2,1,0)

    return data_rho, data_v

# endregion

# region SimpleITK interop.


def read_sitk(path):
    import SimpleITK as sitk
    path = os.path.abspath(path)
    logging.debug('Reading "{}".'.format(path))
    return sitk.ReadImage(path)


def save_sitk(img_sitk, path):
    import SimpleITK as sitk
    path = os.path.abspath(path)
    logging.debug('Saving "{}".'.format(path))
    sitk.WriteImage(img_sitk, path, False)


def save_sitk_as_vti(img_sitk, path, array_name='data'):
    vti = sitk_to_vti(img_sitk, array_name)
    save_vti(vti, path)


def sitk_to_vti(img, array_name, array_type=None):
    import SimpleITK as sitk
    vti = vtk.vtkImageData()
    vti.SetSpacing(img.GetSpacing())
    vti.SetOrigin(img.GetOrigin())
    vti.SetDimensions(img.GetSize())

    voxels = sitk.GetArrayFromImage(img)
    arr = numpy_support.numpy_to_vtk(num_array=voxels.ravel(), deep=True, array_type=array_type)
    arr.SetName(array_name)
    vti.GetPointData().SetScalars(arr)

    return vti


def vti_to_sitk(vti, array):
    import SimpleITK as sitk
    arr_np = vti_to_np(vti, array)
    arr_sitk = sitk.GetImageFromArray(arr_np)
    arr_sitk.SetOrigin(vti.GetOrigin())
    arr_sitk.SetSpacing(vti.GetSpacing())
    return arr_sitk

# endregion

# region Some strange functions


def probe_vtu(vtu_file='output.vtu', point_data=[[0.050640, 0.027959, 0.05213]], fname=None, shape2d=None,
              verbose=False):
    """
    get values of interpolated from vtu mesh on the set of points (N,3)
    """
    vtu = read_vtk(vtu_file)

    points = vtk.vtkPoints()
    for point in point_data:
        points.InsertNextPoint(*point)

    polydata = vtk.vtkPolyData()
    polydata.SetPoints(points)

    probe = vtk.vtkProbeFilter()
    probe.SetSourceData(vtu)
    probe.SetInputData(polydata)
    probe.Update()

    out = probe.GetOutput()
    # out.GetBounds()
    # to get points: numpy_support.vtk_to_numpy(out.GetPoints().GetData()).shape
    pd = out.GetAttributesAsFieldData(0)
    log = logging.getLogger(vtu_file)
    if fname is None:
        # all fields
        output = dict()
        for i in range(pd.GetNumberOfArrays()):
            v_interp_on_grid = numpy_support.vtk_to_numpy(pd.GetArray(i))
            if shape2d:
                v_interp_on_grid = v_interp_on_grid.reshape(shape2d)
            log.debug(("appending in output:", pd.GetArrayName(i)))
            output[pd.GetArrayName(i)] = v_interp_on_grid
        assert (len(output) > 0)
        return output
    else:
        field_numbers = [i for i in range(pd.GetNumberOfArrays()) if pd.GetArrayName(i) == fname]
        assert (len(field_numbers) == 1)
        log.debug(("output:", pd.GetArrayName(field_numbers[0])))
        v_interp_on_grid = numpy_support.vtk_to_numpy(pd.GetArray(field_numbers[0]))
        if shape2d:
            return v_interp_on_grid.reshape(shape2d)
        return v_interp_on_grid


def probe_vti(vti_file='output.vti', point_data=[[0.050640, 0.027959, 0.05213]], fname=None, shape2d=None,
              verbose=False):
    """
    get values of interpolated from vti file mesh on the set of points (N,3)
    """
    vti = read_vtk(vti_file)

    points = vtk.vtkPoints()
    for point in point_data:
        points.InsertNextPoint(*point)

    polydata = vtk.vtkPolyData()
    polydata.SetPoints(points)

    probe = vtk.vtkProbeFilter()
    probe.SetSourceData(vti)
    probe.SetInputData(polydata)
    probe.Update()

    out = probe.GetOutput()
    # out.GetBounds()
    # to get points: numpy_support.vtk_to_numpy(out.GetPoints().GetData()).shape
    pd = out.GetAttributesAsFieldData(0)
    log = logging.getLogger(vti_file)
    if fname is None:
        # all fields
        output = dict()
        for i in range(pd.GetNumberOfArrays()):
            v_interp_on_grid = numpy_support.vtk_to_numpy(pd.GetArray(i))
            if shape2d:
                v_interp_on_grid = v_interp_on_grid.reshape(shape2d)
            log.debug(("appending in output:", pd.GetArrayName(i)))
            output[pd.GetArrayName(i)] = v_interp_on_grid
        assert (len(output) > 0)
        return output
    else:
        field_numbers = [i for i in range(pd.GetNumberOfArrays()) if pd.GetArrayName(i) == fname]
        assert (len(field_numbers) == 1)
        log.debug(("output:", pd.GetArrayName(field_numbers[0])))
        v_interp_on_grid = numpy_support.vtk_to_numpy(pd.GetArray(field_numbers[0]))
        if shape2d:
            return v_interp_on_grid.reshape(shape2d)
        return v_interp_on_grid


def mod_mesh(du, stlfile='c0006.stl', output_fn='surface.vtp',
             write=False, sign=1):
    """
    Move stl in normal direction
    """
    stl = read_vtk(stlfile)
    vertices = numpy_support.vtk_to_numpy(stl.GetPoints().GetData())
    # indices = numpy_support.vtk_to_numpy(stl.GetPolys().GetData()).reshape(-1, 4)[:, 1:4]

    merged = vtk.vtkPolyData()
    merged.DeepCopy(stl)

    # Compute normals to vertices
    normalGenerator = vtk.vtkPolyDataNormals()
    normalGenerator.SetInputData(merged)
    normalGenerator.ComputePointNormalsOn()
    normalGenerator.ComputeCellNormalsOff()
    normalGenerator.SetSplitting(0)
    normalGenerator.SetConsistency(0)
    normalGenerator.Update()

    merged = normalGenerator.GetOutput()
    normals = numpy_support.vtk_to_numpy(merged.GetPointData().GetNormals())

    points = vtk.vtkPoints()

    for normal, pos in zip(normals, vertices):
        points.InsertNextPoint(pos + normal * (-sign * du))

    merged.SetPoints(points)

    return merged


def probe_at_dx(du=0.001, velocity_file=None, stlfile='c0006.stl', output_fn='surface.vtp',
                velocity_name='v [m/s]',
                write=False, mu=1.0, move_mesh=False, sign=1):
    """
    Equidistant points from stl in normal direction
    """
    stl = read_vtk(stlfile)
    vertices = numpy_support.vtk_to_numpy(stl.GetPoints().GetData())
    # indices = numpy_support.vtk_to_numpy(stl.GetPolys().GetData()).reshape(-1, 4)[:, 1:4]
    merged = vtk.vtkPolyData()
    merged.DeepCopy(stl)

    vel_data = read_vtk(velocity_file)

    # Compute normals to vertices
    normalGenerator = vtk.vtkPolyDataNormals()
    normalGenerator.SetInputData(merged)
    normalGenerator.ComputePointNormalsOn()
    normalGenerator.ComputeCellNormalsOff()
    normalGenerator.SetSplitting(0)
    normalGenerator.SetConsistency(0)
    normalGenerator.Update()

    merged = normalGenerator.GetOutput()
    normals = numpy_support.vtk_to_numpy(merged.GetPointData().GetNormals())

    points = vtk.vtkPoints()
    pointsPolyData = vtk.vtkPolyData()

    for normal, pos in zip(normals, vertices):
        points.InsertNextPoint(pos + normal * (-sign * du))

    pointsPolyData.SetPoints(points)
    probe_filter = vtk.vtkProbeFilter()
    probe_filter.SetInputData(pointsPolyData)
    probe_filter.SetSourceData(vel_data)
    probe_filter.GetOutputPort()
    probe_filter.Update()

    probed_data = probe_filter.GetOutput().GetPointData()

    if isinstance(velocity_name, str):
        v_vec = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name))
    elif isinstance(velocity_name, list):
        vx = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name[0]))
        vy = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name[1]))
        vz = numpy_support.vtk_to_numpy(probed_data.GetArray(velocity_name[2]))
        v_vec = np.stack((vx, vy, vz), 1).reshape((-1, 3))
    else:
        raise NotImplementedError

    logging.debug(v_vec.shape)

    velocity = vtk.vtkFloatArray()
    velocity.SetNumberOfComponents(1)
    velocity.SetName("X_at_epsilon")

    for v in v_vec:
        # if np.max(v) > 1e33 or np.min(v) < -1e33:
        #     s = np.array([0.00])

        velocity.InsertNextTypedTuple([v])

    merged.GetPointData().AddArray(velocity)

    if move_mesh:
        merged.SetPoints(points)
        logging.debug('moving point by dx inside')

    merged.Modified()
    if write:
        save_vtp(merged, output_fn)

    return merged


def scale_and_trans(vtk_data=None, output=None,
                    scale=1000.0,
                    deltaxyz=[-14.52326308, 180.637182, 161.81502267]):
    """
    Performs scaling (e.g from meters to mm) and translation of the dataset.
    Note that `vtk_data` is reader.GetOutput()
    """
    transform = vtk.vtkTransform()
    transform.Scale(scale, scale, scale)
    transform.Translate(*deltaxyz)
    transformFilter = vtk.vtkTransformFilter()
    transformFilter.SetTransform(transform)
    transformFilter.SetInputData(vtk_data)
    transformFilter.Update()
    if output is None:
        return transformFilter.GetOutput()
    else:
        save_vtu(transformFilter.GetOutput(), output)

# endregion



import numpy as np


class Chunk(np.ndarray):
    def __new__(cls, input_array, origin=None, spacing=None):
        obj = np.asarray(input_array).view(cls)
        obj.origin = origin
        obj.spacing = spacing
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.origin = getattr(obj, 'origin', None)
        self.spacing = getattr(obj, 'spacing', None)

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, val):
        self._spacing = val

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, val):
        self._origin = val

    def to_dict(self):
        return {
            b'origin': self.origin,
            b'spacing': self.spacing,
            b'ndarray': self
        }

    @staticmethod
    def from_dict(d):
        return Chunk(d[b'ndarray'].copy(), origin=d[b'origin'], spacing=d[b'spacing'])



import os
import unittest

import numpy as np

from chunky3d import Sparse, point_probe
from chunky3d.sparse_func import contour
from chunky3d.vtk_utils import read_vtk, save_stl, save_vtp, save_vti, vti_to_np, add_np_to_vti


class TestContour(unittest.TestCase):
    def test_contour(self):
        # import warnings
        # warnings.filterwarnings('error')
        shape = (100, 100, 100)
        sp = Sparse(shape=shape)
        step = 0

        step += 1
        sp[50, 50, 50] = 100
        c = contour(sp, 100)
        self.assertEqual(c.GetNumberOfCells(), 0)
        self.assertEqual(c.GetNumberOfPoints(), 0)
        #save_vtp(c,f"test_contour_step{step}_{'_'.join([str(s) for s in shape])}.vtp")

        step+=1
        sp[50, 50, 50] = 101
        c = contour(sp, 100)
        self.assertEqual(c.GetNumberOfPoints(), 6)
        self.assertEqual(c.GetNumberOfCells(), 8)
        #save_vtp(c,f"test_contour_step{step}_{'_'.join([str(s) for s in shape])}.vtp")


if __name__ == '__main__':
    unittest.main()